#!/usr/bin/env python
# coding: utf-8
# Copyright (c) 2011 NEXT-TELL
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# This code is based on code Copyright (c) 2010, Facebook, Inc.
# 
import os
# dummy config to enable registering django template filters
os.environ['DJANGO_SETTINGS_MODULE'] = u'conf'
from google.appengine.dist import use_library
use_library('django', '1.2')

from django.template.defaultfilters import register
from django.utils import simplejson as json
from functools import wraps
from google.appengine.api import urlfetch, taskqueue
from google.appengine.ext import db, webapp
from google.appengine.ext.webapp import util, template
from google.appengine.runtime import DeadlineExceededError
from random import randrange
from uuid import uuid4
import Cookie
import base64
import cgi
import conf
import datetime
import hashlib
import hmac
import logging
import math
import pickle
import random
import time
import traceback
import urllib
#             0      1       2      3             4               5               6               7
ELEMENTS = ['search engine marketing', 'social media marketing', 'viral marketing', 'flash mob marketing', 'mobile marketing', 'augmented reality marketing', 'in-game advertising', 'advergaming']
TRIADS = [[7,6,3], [1,0,4], [5,2,6], [0,3,5], [2,1,4]]

def htmlescape(text):
    """Escape text for use as HTML"""
    return cgi.escape(
        text, True).replace("'", '&#39;').encode('ascii', 'xmlcharrefreplace')


@register.filter(name=u'get_name')
def get_name(dic, index):
    """Django template filter to render name"""
    return dic[index].name


@register.filter(name=u'get_picture')
def get_picture(dic, index):
    """Django template filter to render picture"""
    return dic[index].picture

@register.simple_tag
def get_element(list, index):
# You should catch IndexError here.
    return list[index]

def select_random(lst, limit):
    """Select a limited set random non Falsy values from a list"""
    final = []
    size = len(lst)
    while limit and size:
        index = randrange(min(limit, size))
        size = size - 1
        elem = lst[index]
        lst[index] = lst[size]
        if elem:
            limit = limit - 1
            final.append(elem)
    return final


_USER_FIELDS = u'name,email,picture,friends'
class User(db.Model):
    user_id = db.StringProperty(required=True)
    access_token = db.StringProperty(required=True)
    name = db.StringProperty(required=True)
    picture = db.StringProperty(required=True)
    email = db.StringProperty()
    friends = db.StringListProperty()
    friendnames = db.StringListProperty()
    dirty = db.BooleanProperty()
    created = db.DateTimeProperty()
    updated = db.DateTimeProperty()
    
    def refresh_data(self):
        """Refresh this user's data using the Facebook Graph API"""
        me = Facebook().api(u'/me',
            {u'fields': _USER_FIELDS, u'access_token': self.access_token})
        self.dirty = False
        self.name = me[u'name']
        self.email = me.get(u'email')
        self.picture = me[u'picture']
        self.friends = [user[u'id'] for user in me[u'friends'][u'data']]
        self.friendnames = [user[u'name'] for user in me[u'friends'][u'data']]
        self.updated = datetime.datetime.now()
        return self.put()

# from http://stackoverflow.com/questions/2028355/datastore-list-of-lists
class GenericListProperty(db.Property):
    data_type = db.Blob
    def validate(self, value):
        if type(value) is not list:
            raise db.BadValueError('%s Property %s must be a list, not %s.' % (self.name, type(value), value))
        return value

    def get_value_for_datastore(self, model_instance):
        return db.Blob(pickle.dumps(getattr(model_instance,self.name)))

    def make_value_from_datastore(self, value):
        return pickle.loads(value)
        
class Grid(db.Model):
    grid_version = db.StringProperty(default='2-0-0')
    user_id = db.StringProperty(required=True)
    date = db.DateTimeProperty(required=True)
    topic = db.StringProperty()
    elements = db.StringListProperty()
    constructs_left = db.StringListProperty()
    constructs_right = db.StringListProperty()
    scores = db.ListProperty(int)
    svgs = db.ListProperty(db.Text);
    svgwhens = db.ListProperty(datetime.datetime)
    reflections = db.StringListProperty()
    reflectionwhens = db.ListProperty(datetime.datetime)
    triads = GenericListProperty();
    triads_repr = db.StringProperty()
    nexttriad = db.IntegerProperty()
    
    maxdiff = None
    # return [(1,"foo"),(5,"bar"),(0,"baz")]
    def getTriad(self):
        if self.triads:
            triadindexes = self.triads[self.nexttriad]
            triad = list();
            for index in triadindexes:
                triad.append((index,self.elements[index]))
            #self.nexttriad = self.nexttriad + 1
            # only put if no error...
            #self.put()
            return triad
        else:
            return random.sample(list(enumerate(self.elements)),3)

    def getMaxDiff(self):
        return self.maxdiff

    def getDiffs(self):
        diffs = list();
        maxdiff = 0;
        for index1 in range(len(self.elements)-1):
            for index2 in range(index1,len(self.elements)):
                total = 0.0
                for index3 in range(0,len(self.constructs_left)):
                    diff = self.scores[index3 * len(self.elements) + (index1)] - self.scores[index3 * len(self.elements) + (index2)]
                    total += diff*diff
                realdiff = math.sqrt(total)
                if realdiff > maxdiff:
                    maxdiff = realdiff
                diffs.append((index1, index2, realdiff))
        self.maxdiff = maxdiff
        return diffs
        
    @staticmethod
    def find_by_user_ids(user_ids, limit=50):
        if user_ids:
            return Grid.gql(u'WHERE user_id IN :1 ORDER BY date DESC', user_ids).fetch(limit)
        else:
            return []

class TriadList(db.Model):
    triads = GenericListProperty()
    triads2 = db.StringProperty()
    @staticmethod
    def getTriadList():
        try:
            t = db.GqlQuery("SELECT * FROM TriadList")
            return t[0].triads
        except Exception, e: 
            return TRIADS
    @staticmethod
    def setTriadList(triadString):
        triadlist = TriadList(triads = TRIADS)            
        try:
            triadlist2 = triadString
            triadlist = TriadList(triads = eval(triadString),triads2 = triadString)
        except Exception, e:
            triadlist = TriadList(triads = TRIADS)
        q = db.GqlQuery("SELECT __key__ from TriadList")
        results = q.fetch(100)
        db.delete(results)
        triadlist.put();

class Elements(db.Model):
    elements = GenericListProperty()
    elements2 = db.StringProperty()

    @staticmethod
    def getElements():
        try:
            t = db.GqlQuery("SELECT * FROM Elements")
            return t[0].elements
        except Exception, e: 
            return ELEMENTS
    @staticmethod
    def setElements(elementList):
        elements = Elements(elements = elementList)            
        try:
            #elementList = ['one','two','three','four']
            elements2 = repr(elementList)
            elements = Elements(elements = elementList,elements2=elements2)
        except Exception, e:
            elements = Elements(elements = ELEMENTS)
            
        # delete old elements
        q = db.GqlQuery("SELECT __key__ from Elements")
        results = q.fetch(100)
        db.delete(results)
        
        elements.put();

class Configs(db.Model):
    elementSingular = db.StringProperty()
    elementPlural = db.StringProperty()
    introblurb = db.TextProperty()
    thinkof = db.TextProperty()
    similarphrase = db.TextProperty()
    differentphrase = db.TextProperty()
    wallid = db.StringProperty()

    @staticmethod
    def setConfigs(elementSingular,elementPlural,introblurb,thinkof,similarphrase,differentphrase,nowrate,wallid):
        #         elementSingular = "thing"
        #         elementPlural = "things"
        #         introblurb = """
        #         In this exercise you will be asked to think about the relationships between consumer decision making processes for some types of product. You will go through a series of prompts that will select three of those products at a time and ask you to identify similarities and differences in how you think consumers make decisions about purchasing them.  You will do this a total of 10 times.
        # <p>
        #         When you are done you will be shown several representations of the relationships you identified between the products.
        #         You will also be given the option of posting the results to the course wall on Facebook.
        #       <p>
        #       <b>Please note:</b>  you should use Internet Explorer 9 or newer, Firefox 4 or newer, Chrome 10 or newer, or Opera 11 or newer if you want to be able to see all the different representations of the relationships.
        #       <p>
        #       The product types are:
        #       """
        #       thinkof = """
        #         Think of the following three products from the perspective of consumer decision making.
        #         <p>How is the consumer decision making process for <b>two of them similar</b> and <b>different from the third?</b></p>
        #         <p>Select the <b>one for which the consumer decision making process is different</b>
        #         <p>
        #         """
        #         similarphrase = """
        #         Enter a phrase characterizing how the <b>consumer decision making processes</b> for the other two products are <b>similar</b>
        #         """
        #         differentphrase = """
        #         Enter a phrase characterizing how the <b>consumer decision making process</b> for the selected product is <b>different</b>
        #         """
        #         nowrate = """
        #         Now rate the consumer decision making process for each product on the construct
        #         """
        c = Configs(elementSingular = elementSingular,
            elementPlural = elementPlural,
            introblurb = introblurb,
            thinkof = thinkof,
            similarphrase = similarphrase,
            differentphrase = differentphrase,
            nowrate = nowrate,
            wallid = wallid
            )
        # delete old elements
        q = db.GqlQuery("SELECT __key__ from Configs")
        results = q.fetch(100)
        db.delete(results)

        c.put();

    @staticmethod
    def getConfigs():
        try:
            t = db.GqlQuery("SELECT * FROM Configs")
            return t[0]
        except Exception, e: 
            pass
            
class Run(db.Model):
    user_id = db.StringProperty(required=True)
    location = db.StringProperty(required=True)
    distance = db.FloatProperty(required=True)
    date = db.DateProperty(required=True)
    
    @staticmethod
    def find_by_user_ids(user_ids, limit=50):
        if user_ids:
            return Run.gql(u'WHERE user_id IN :1', user_ids).fetch(limit)
        else:
            return []
    
    @property
    def pretty_distance(self):
        return u'%.2f' % self.distance


class RunException(Exception):
    pass


class FacebookApiError(Exception):
    def __init__(self, result):
        self.result = result
    
    def __str__(self):
        return self.__class__.__name__ + ': ' + json.dumps(self.result)


class Facebook(object):
    """Wraps the Facebook specific logic"""
    def __init__(self, app_id=conf.FACEBOOK_APP_ID,
            app_secret=conf.FACEBOOK_APP_SECRET):
        self.app_id = app_id
        self.app_secret = app_secret
        self.user_id = None
        self.access_token = None
        self.signed_request = {}
    
    def api(self, path, params=None, method=u'GET', domain=u'graph'):
        """Make API calls"""
        if not params:
            params = {}
        params[u'method'] = method
        if u'access_token' not in params and self.access_token:
            params[u'access_token'] = self.access_token
        result = json.loads(urlfetch.fetch(
            url=u'https://' + domain + u'.facebook.com' + path,
            payload=urllib.urlencode(params),
            method=urlfetch.POST,
            headers={
                u'Content-Type': u'application/x-www-form-urlencoded'})
            .content)
        if isinstance(result, dict) and u'error' in result:
            raise FacebookApiError(result)
        return result
    
    def load_signed_request(self, signed_request):
        """Load the user state from a signed_request value"""
        try:
            sig, payload = signed_request.split(u'.', 1)
            sig = self.base64_url_decode(sig)
            data = json.loads(self.base64_url_decode(payload))
            
            expected_sig = hmac.new(
                self.app_secret, msg=payload, digestmod=hashlib.sha256).digest()
            
            # allow the signed_request to function for upto 1 day
            if sig == expected_sig and \
                    data[u'issued_at'] > (time.time() - 86400):
                self.signed_request = data
                self.user_id = data.get(u'user_id')
                self.access_token = data.get(u'oauth_token')
        except ValueError, ex:
            pass # ignore if can't split on dot
    
    @property
    def user_cookie(self):
        """Generate a signed_request value based on current state"""
        if not self.user_id:
            return
        payload = self.base64_url_encode(json.dumps({
            u'user_id': self.user_id,
            u'issued_at': str(int(time.time())),
        }))
        sig = self.base64_url_encode(hmac.new(
            self.app_secret, msg=payload, digestmod=hashlib.sha256).digest())
        return sig + '.' + payload
    
    @staticmethod
    def base64_url_decode(data):
        data = data.encode(u'ascii')
        data += '=' * (4 - (len(data) % 4))
        return base64.urlsafe_b64decode(data)
    
    @staticmethod
    def base64_url_encode(data):
        return base64.urlsafe_b64encode(data).rstrip('=')


class CsrfException(Exception):
    pass


class BaseHandler(webapp.RequestHandler):
    facebook = None
    user = None
    csrf_protect = True
    
    def initialize(self, request, response):
        """General initialization for every request"""
        super(BaseHandler, self).initialize(request, response)
        
        try:
            self.init_facebook()
            self.init_csrf()
            self.response.headers[u'P3P'] = u'CP=HONK'  # iframe cookies in IE
        except Exception, ex:
            self.log_exception(ex)
            raise
    
    def handle_exception(self, ex, debug_mode):
        """Invoked for unhandled exceptions by webapp"""
        self.log_exception(ex)
        self.render(u'error',
            trace=traceback.format_exc(), debug_mode=debug_mode)
    
    def log_exception(self, ex):
        """Internal logging handler to reduce some App Engine noise in errors"""
        msg = ((str(ex) or ex.__class__.__name__) +
                u': \n' + traceback.format_exc())
        if isinstance(ex, urlfetch.DownloadError) or \
           isinstance(ex, DeadlineExceededError) or \
           isinstance(ex, CsrfException) or \
           isinstance(ex, taskqueue.TransientError):
            logging.warn(msg)
        else:
            logging.error(msg)
    
    def set_cookie(self, name, value, expires=None):
        """Set a cookie"""
        if value is None:
            value = 'deleted'
            expires = datetime.timedelta(minutes=-50000)
        jar = Cookie.SimpleCookie()
        jar[name] = value
        jar[name]['path'] = u'/'
        if expires:
            if isinstance(expires, datetime.timedelta):
                expires = datetime.datetime.now() + expires
            if isinstance(expires, datetime.datetime):
                expires = expires.strftime('%a, %d %b %Y %H:%M:%S')
            jar[name]['expires'] = expires
        self.response.headers.add_header(*jar.output().split(u': ', 1))
    
    def render(self, name, **data):
        """Render a template"""
        if not data:
            data = {}
        data[u'js_conf'] = json.dumps({
            u'appId': conf.FACEBOOK_APP_ID,
            u'canvasName': conf.FACEBOOK_CANVAS_NAME,
            u'userIdOnServer': self.user.user_id if self.user else None,
        })
        data[u'logged_in_user'] = self.user
        data[u'message'] = self.get_message()
        data[u'csrf_token'] = self.csrf_token
        data[u'canvas_name'] = conf.FACEBOOK_CANVAS_NAME
        self.response.out.write(template.render(
            os.path.join(
                os.path.dirname(__file__), 'templates', name + '.html'),
            data))
    
    def init_facebook(self):
        """Sets up the request specific Facebook and User instance"""
        facebook = Facebook()
        user = None
        
        # initial facebook request comes in as a POST with a signed_request
        if u'signed_request' in self.request.POST:
            facebook.load_signed_request(self.request.get('signed_request'))
            # we reset the method to GET because a request from facebook with a
            # signed_request uses POST for security reasons, despite it
            # actually being a GET. in webapp causes loss of request.POST data.
            self.request.method = u'GET'
            self.set_cookie(
                'u', facebook.user_cookie, datetime.timedelta(minutes=1440))
        elif 'u' in self.request.cookies:
            facebook.load_signed_request(self.request.cookies.get('u'))
        
        # try to load or create a user object
        if facebook.user_id:
            user = User.get_by_key_name(facebook.user_id)
            if user:
                # update stored access_token
                if facebook.access_token and \
                        facebook.access_token != user.access_token:
                    user.access_token = facebook.access_token
                    user.put()
                # refresh data if we failed in doing so after a realtime ping
                # or if we don't have the friendname field loaded...
                if user.dirty or len(user.friendnames) != len(user.friends):
                    user.refresh_data()
                # restore stored access_token if necessary
                if not facebook.access_token:
                    facebook.access_token = user.access_token
            
            if not user and facebook.access_token:
                me = facebook.api(u'/me', {u'fields': _USER_FIELDS})
                try:
                    friends = [user[u'id'] for user in me[u'friends'][u'data']]
                    friendnames = [user[u'name'] for user in me[u'friends'][u'data']]
                    user = User(key_name=facebook.user_id,
                        user_id=facebook.user_id, friends=friends, friendnames=friendnames, 
                        created = datetime.datetime.now(), updated = datetime.datetime.now(),
                        access_token=facebook.access_token, name=me[u'name'],
                        email=me.get(u'email'), picture=me[u'picture'])
                    user.put()
                except KeyError, ex:
                    pass # ignore if can't get the minimum fields
        
        self.facebook = facebook
        self.user = user
    
    def init_csrf(self):
        """Issue and handle CSRF token as necessary"""
        self.csrf_token = self.request.cookies.get(u'c')
        if not self.csrf_token:
            self.csrf_token = str(uuid4())[:8]
            self.set_cookie('c', self.csrf_token)
        if self.request.method == u'POST' and self.csrf_protect and \
                self.csrf_token != self.request.POST.get(u'_csrf_token'):
            raise CsrfException(u'Missing or invalid CSRF token.')
    
    def set_message(self, **obj):
        """Simple message support"""
        self.set_cookie('m', base64.b64encode(json.dumps(obj)) if obj else None)
    
    def get_message(self):
        """Get and clear the current message"""
        message = self.request.cookies.get(u'm')
        if message:
            self.set_message()  # clear the current cookie
            return json.loads(base64.b64decode(message))


def user_required(fn):
    """Decorator to ensure a user is present"""
    @wraps(fn)
    def wrapper(*args, **kwargs):
        handler = args[0]
        if handler.user:
            return fn(*args, **kwargs)
        handler.redirect(u'/')
    return wrapper

class RecentGridsHandler(BaseHandler):
    """Show recent grids for the user and friends"""
    def get(self):
        if self.user:
            friends = {}
            elements = Elements.getElements(); # CPT
            introblurb = Configs.getConfigs().introblurb
            #             """
            #             In this exercise you will be asked to think about the relationships between consumer decision making processes for some types of product. You will go through a series of prompts that will select three of those products at a time and ask you to identify similarities and differences in how you think consumers make decisions about purchasing them.  You will do this a total of 10 times.
            # <p>
            #             When you are done you will be shown several representations of the relationships you identified between the products.
            #             You will also be given the option of posting the results to the course wall on Facebook.
            # <p>
            # <b>Please note:</b>  you should use Internet Explorer 9 or newer, Firefox 4 or newer, Chrome 10 or newer, or Opera 11 or newer if you want to be able to see all the different representations of the relationships.
            # <p>
            # The product types are:
            # """
            # 
            self.render(u'grids',
                topic = u'online marketing v3',
                elements=elements,
                friends=friends,
                introblurb=introblurb
                # user_recent_grids=Grid.find_by_user_ids(
                #     [self.user.user_id], limit=5),
                # friends_grids=Grid.find_by_user_ids(friends.keys()),
            )
        else:
            self.render(u'welcome')


class UserGridsHandler(BaseHandler):
    """Show a specific user's grids, ensure friendship with the logged in user"""
    @user_required
    def get(self, user_id):
        if self.user.friends.count(user_id) or self.user.user_id == user_id:
            user = User.get_by_key_name(user_id)
            if not user:
                self.set_message(type=u'error',
                    content=u'That user does not use FARGO.')
                self.redirect(u'/')
                return
            
            self.render(u'user',
                user=user,
                grids=Grid.find_by_user_ids([user_id]),
            )
        else:
            self.set_message(type=u'error',
                content=u'You are not allowed to see that.')
            self.redirect(u'/')

class TriadHandler(BaseHandler):
    """Show a triad and elicit the differences"""
    @user_required
    def get(self):
        try:
            grids=Grid.find_by_user_ids([self.user.user_id], limit=1)
            grid = grids[0]
            user_triad = grid.getTriad()

            self.render(u'triads',
                current = grid.nexttriad+1,
                total = len(grid.triads),
                user_triad = user_triad,
                topic = grid.topic,
                thinkof = Configs.getConfigs().thinkof,
                differentphrase = Configs.getConfigs().differentphrase,
                similarphrase = Configs.getConfigs().similarphrase
            )

        except Exception, e:
            self.set_message(type=u'error',
                content=u'Unknown error occured. (' + unicode(e) + u') ' + unicode(traceback.print_exc()))
            
        
class RatingsHandler(BaseHandler):
    @user_required
    def post(self):
        try:
            grids=Grid.find_by_user_ids([self.user.user_id], limit=1)
            grid = grids[0] 

            choiceIndex = self.request.POST[u'choice']
            different = self.request.POST[u'different']
            alike = self.request.POST[u'alike']
            if not len(different):
                self.set_message(type=u'error',
                    content=u'The phrase describing how the selected product is different cannot be empty.') 
                self.redirect(u'/triad')
                return
            if not len(alike):
                self.set_message(type=u'error',
                    content=u'The phrase describing how the other products are alike cannot be empty.') 
                self.redirect(u'/triad')
                return
            if alike == different:
                self.set_message(type=u'error',
                    content=u'The phrase describing differences should not be the same as the one describing similarities.') 
                self.redirect(u'/triad')
                return
            grid.constructs_left.append(different)
            grid.constructs_right.append(alike)
            grid.put()
            elementIndices = self.request.POST.getall(u'elementIndices')
            defaults = list()
            for i in range(0,len(grid.elements)):
                v = 0
                if i == int(choiceIndex):
                    v = 1
                else:
                    for elementIndex in elementIndices:
                        if i == int(elementIndex):
                            v=5
                l = list()
                l.append(i)
                l.append(grid.elements[i])
                l.append(v)
                defaults.append(l)
            nowrate = """
            Now rate the consumer decision making process for each product on the construct
            """
            self.render(u'triadratings',
                current = grid.nexttriad+1,
                total = len(grid.triads),
                choiceIndex = choiceIndex,
                different = different,
                alike = alike,
                triadIndices = elementIndices,
                elements = list(enumerate(grid.elements)),
                nowrate = nowrate,
                defaults = defaults
            )
            grid.nexttriad = grid.nexttriad + 1
            grid.put() # only update if no error...
        except KeyError:
            self.set_message(type=u'error',
                content=u'You need to select a product that is different from the other two.') 
            self.redirect(u'/triad')   
        except Exception, e:
            self.set_message(type=u'error',
                content=u'Unknown error occured. (' + unicode(e) + u') ' + unicode(traceback.print_exc()))

class SaveRatingsHandler(BaseHandler):
    @user_required
    def post(self):
        try:
            grids=Grid.find_by_user_ids([self.user.user_id], limit=1)
            grid = grids[0]
            scores = grid.scores
            for score in range(0,len(grid.elements)):
                scores.append(int(self.request.POST[str(score)]))
            grid.scores = scores
            grid.put()
            # if enough, then show results... otherwise redirect to more elicitation
#            if len(grid.constructs_left) > 4: # CPT
            if grid.nexttriad >= len(grid.triads):
                self.redirect(u'/results')
            else:
                self.redirect(u'/triad')
        except Exception, e:
            self.set_message(type=u'error',
                content=u'Unknown error occured. (' + unicode(e) + u') ' + unicode(traceback.print_exc())) 


class GridHandler(BaseHandler):
    """Add a run"""
    @user_required
    def post(self):
        try:
            topic = self.request.POST[u'topic'].strip()
            if not topic:
                raise RunException(u'Please specify a topic.')
            
            # distance = float(self.request.POST[u'distance'].strip())
            # if distance < 0:
            #     raise RunException(u'Invalid distance.')
            
            # date_year = int(self.request.POST[u'date_year'].strip())
            # date_month = int(self.request.POST[u'date_month'].strip())
            # date_day = int(self.request.POST[u'date_day'].strip())
            # if date_year < 0 or date_month < 0 or date_day < 0:
            #     raise RunException(u'Invalid date.')
            # date = datetime.date(date_year, date_month, date_day)
            
            # run = Run(
            #     user_id=self.user.user_id,
            #     location=location,
            #     distance=distance,
            #     date=datetime.now(),
            # )
            # elements = self.request.get('elements')
            # elementlist = elements.split('\r\n')
            elementlist = Elements.getElements() # CPT
            nonemptyelements = list()
            for element in elementlist:
                cleanElement = element.strip()
                if len(cleanElement) > 0:
                    nonemptyelements.append(element)
            #triads = random.sample(TRIADS,2)
            triads = TriadList.getTriadList()
            random.shuffle(triads)
            grid = Grid(
                user_id = self.user.user_id,
                topic = topic,
                date = datetime.datetime.now(),
                elements = nonemptyelements,
                # triad index of elements, which start at 0
                triads = triads,
                triads_repr = repr(triads), # this makes it easy to see what's going on
                nexttriad = 0
            )
            # list of tuples?

            grid.put()
# The NEXT-TELL group: 132267966825411
# Online Marketing group: 133258548012
# CSSL group: 220154214670167
# myself = /me/feed
            #mex = Facebook().api(u'/me/feed', {u'message': u'Added a grid about ' + unicode(topic) + u' with ' + unicode(len(elementlist)) + u' elements.', u'access_token': self.user.access_token}, u'POST')
            # title = topic
            # publish = u'<a onclick=\'publishRun(' + \
            #         json.dumps(htmlescape(title)) + u')\'>Post to facebook.</a>'
            # self.set_message(type=u'success',
            #     content=u'Added your grid. ')
        except RunException, e:
            self.set_message(type=u'error', content=unicode(e))
        except KeyError:
            self.set_message(type=u'error',
                content=u'Please specify location, distance & date.')
        except ValueError:
            self.set_message(type=u'error',
                content=u'Please specify a valid distance & date.')
        except Exception, e:
            self.set_message(type=u'error',
                content=u'Unknown error occured. (' + unicode(e) + u') ' + unicode(traceback.print_exc()))
        self.redirect(u'/triad')

class Line:
    construct_left = '';
    construct_right = '';
    scores = list();
    triads = ''
    def __init__(self,cleft,cright,ss,t):
        self.construct_left = cleft
        self.construct_right = cright
        self.scores = ss
        self.triads = t

class ResultsHandler(BaseHandler):
    @user_required
    def get(self):
        grids=Grid.find_by_user_ids([self.user.user_id], limit=1)
        grid = grids[0]
        lines = list();
        for construct in range(0,len(grid.triads)):
            
            construct_left = grid.constructs_left[construct]
            construct_right = grid.constructs_right[construct]
            scores = list()
            for element in range(0,len(grid.elements)):
                scores.append(grid.scores[construct*len(grid.elements)  + element])
            line = Line(construct_left,construct_right,scores,'')
            lines.append(line)
        self.render(u'results',
            elements = grid.elements,
            lines = lines
        )

class MiserablesHandler(BaseHandler):
    @user_required
    def get(self):
        grids=Grid.find_by_user_ids([self.user.user_id], limit=1)
        grid = grids[0]
        diffs = grid.getDiffs();
        maxdiff = grid.getMaxDiff();
        miserables = "{\"nodes\":[ "      
        for elementIndex in range(len(grid.elements)):
            miserables += "{\"name\":\"" + grid.elements[elementIndex] + "\", \"group\":1}"
            if elementIndex < len(grid.elements)-1:
                miserables += ","
#            miserables += "\n"
        miserables += "],\"links\":[ "
        first = True
        for diffIndex in range(len(diffs)):
            if diffs[diffIndex][2] < 10:
                if not first:
                    miserables += ","
                miserables += "{\"source\": %d, \"target\": %d, \"value\": %d} " % (diffs[diffIndex][0],diffs[diffIndex][1],maxdiff-diffs[diffIndex][2])
            first = False
        miserables += " ] }"
        self.response.headers['Content-Type'] = "application/json"
        self.response.out.write(miserables)
        
class NetworkHandler(BaseHandler):
    @user_required
    def get(self):
        grids=Grid.find_by_user_ids([self.user.user_id], limit=1)
        grid = grids[0]
        diffs = grid.getDiffs();
        miserables = """
        var miserables = {
          nodes:[     
        """    
        for elementIndex in range(len(grid.elements)):
            miserables += "{nodeName:\"" + grid.elements[elementIndex] + "\", group:1}"
            if elementIndex < len(grid.elements)-1:
                miserables += ","
            miserables += "\n"
        miserables += """
        ],
        links:[
"""
        for diffIndex in range(len(diffs)):
            miserables += "{source: %d, target: %d, value: %d} " % (diffs[diffIndex])
            if diffIndex < len(diffs)-1:
                miserables += ","
            miserables += "\n"
        miserables += """
          ]
        };
"""
        #print miserables;
        lines = list();
        for construct in range(0,len(grid.triads)):
            
            construct_left = grid.constructs_left[construct]
            construct_right = grid.constructs_right[construct]
            scores = list()
            for element in range(0,len(grid.elements)):
                scores.append(grid.scores[construct*len(grid.elements)  + element])
            line = Line(construct_left,construct_right,scores,'')
            lines.append(line)
        self.render(u'network',
            miserables = miserables,
            elements = grid.elements,
            constructs_left = grid.constructs_left,
            constructs_right = grid.constructs_right,
            lines = lines,
            minlinkvalue = 3
        )

class RealtimeHandler(BaseHandler):
    """Handles Facebook Real-time API interctions"""
    csrf_protect = False
    
    def get(self):
        if (self.request.GET.get(u'setup') == u'1' and
            self.user and conf.ADMIN_USER_IDS.count(self.user.user_id)):
            self.setup_subscription()
            self.set_message(type=u'success',
                content=u'Successfully setup Real-time subscription.')
        elif (self.request.GET.get(u'hub.mode') == u'subscribe' and
              self.request.GET.get(u'hub.verify_token') ==
                  conf.FACEBOOK_REALTIME_VERIFY_TOKEN):
            self.response.out.write(self.request.GET.get(u'hub.challenge'))
            logging.info(
                u'Successful Real-time subscription confirmation ping.')
            return
        else:
            self.set_message(type=u'error',
                content=u'You are not allowed to do that.')
        self.redirect(u'/')
    
    def post(self):
        body = self.request.body
        if self.request.headers[u'X-Hub-Signature'] != (u'sha1=' + hmac.new(
            self.facebook.app_secret,
            msg=body,
            digestmod=hashlib.sha1).hexdigest()):
            logging.error(
                u'Real-time signature check failed: ' + unicode(self.request))
            return
        data = json.loads(body)
        
        if data[u'object'] == u'user':
            for entry in data[u'entry']:
                taskqueue.add(url=u'/task/refresh-user/' + entry[u'id'])
                logging.info('Added task to queue to refresh user data.')
        else:
            logging.warn(u'Unhandled Real-time ping: ' + body)
    
    def setup_subscription(self):
        path = u'/' + conf.FACEBOOK_APP_ID + u'/subscriptions'
        params = {
            u'access_token': conf.FACEBOOK_APP_ID + u'|' +
                             conf.FACEBOOK_APP_SECRET,
            u'object': u'user',
            u'fields': _USER_FIELDS,
            u'callback_url': conf.EXTERNAL_HREF + u'realtime',
            u'verify_token': conf.FACEBOOK_REALTIME_VERIFY_TOKEN,
        }
        response = self.facebook.api(path, params, u'POST')
        logging.info(u'Real-time setup API call response: ' + unicode(response))


class RefreshUserHandler(BaseHandler):
    """Used as an App Engine Task to refresh a single user's data if possible"""
    csrf_protect = False
    
    def post(self, user_id):
        logging.info('Refreshing user data for ' + user_id)
        user = User.get_by_key_name(user_id)
        if not user:
            return
        try:
            user.refresh_data()
        except FacebookApiError:
            user.dirty = True
            user.put()

class SaveImageHandler(BaseHandler):
    csrf_protect = False
    @user_required
    def post(self):
        try:
            grids=Grid.find_by_user_ids([self.user.user_id], limit=1)
            grid = grids[0]
            svg = self.request.POST[u'svgString']
            grid.svgs.append(db.Text(svg))
            grid.svgwhens.append(datetime.datetime.now())
            grid.put()
            gridkey = grid.key()
            gridoffset=len(grid.svgs)-1
            
            self.render(u'svg',svgString = svg)            
            
            # The NEXT-TELL group: 132267966825411
            # Online Marketing group: 133258548012
            # CSSL group: 220154214670167
            # myself = /me/feed
            url = u'http://fargobeta.appspot.com/showimage?gridid=%s&gridoffset=%d' % (gridkey, gridoffset)
            posturl = u'/%s/feed' % (Configs.getConfigs().wallid)
            if u'posttowall' in self.request.POST.keys():
                mex = Facebook().api(posturl, 
                    {u'message': u'I added a repertory grid using FARGO.', 
                    u'link': url,
                    u'name': u'My repertory grid',
                    # u'caption': u'caption',
                    # u'description': u'description',
                    u'access_token': self.user.access_token},
                    u'POST')

        except Exception, e:
            pass
            # self.set_message(type=u'error',
            #     content=u'Unknown error occured. (' + unicode(e) + u') ' + unicode(traceback.print_exc()))

class ShowImageHandler(BaseHandler):
    csrf_protect = False
    def get(self):
        try:
            gridid = self.request.GET[u'gridid']
            gridoffset = self.request.GET[u'gridoffset']
            grid = Grid.get(gridid)
            
            lines = list();
            for construct in range(0,len(grid.triads)):

                construct_left = grid.constructs_left[construct]
                construct_right = grid.constructs_right[construct]
                scores = list()
                for element in range(0,len(grid.elements)):
                    scores.append(grid.scores[construct*len(grid.elements)  + element])
                line = Line(construct_left,construct_right,scores,'')
                lines.append(line)
            
            
            
            svg = grid.svgs[int(gridoffset)]

            self.render(u'svgshared',svgString = svg,constructs_left = grid.constructs_left, constructs_right = grid.constructs_right, lines=lines, elements=grid.elements)
        except Exception, e:
            self.set_message(type=u'error',
                content=u'Unknown error occured. (' + unicode(e) + u') ' + unicode(traceback.print_exc()))


class ShowAllHandler(BaseHandler):
    csrf_protect = False
    @user_required
    def get(self):
        if self.user and conf.ADMIN_USER_IDS.count(self.user.user_id):
            try:
                grids = Grid.gql("WHERE topic = :topic",topic = 'online marketing v3')
                elements = None;
                alllines = list()
                allconstructs_left = list()
                allconstructs_right = list()
                allsvgs = list()
                allkeys = list()
                alldurations = list()
                count2 = 0
                for grid in grids:
                    count2 = count2 + 1
                    try:
                        elements = grid.elements
                        triads = grid.triads
                        lines = list();
                        for construct in range(0,len(grid.constructs_left)):
                            construct_left = grid.constructs_left[construct]
                            construct_right = grid.constructs_right[construct]
                            scores = list()
                            for element in range(0,len(grid.elements)):
                                scores.append(grid.scores[construct*len(grid.elements)  + element])
                            triadString = "%s, %s, %s" % (grid.elements[triads[construct][0]],
                                grid.elements[triads[construct][1]],
                                grid.elements[triads[construct][2]])
                            line = Line(construct_left,construct_right,scores, triadString)
                            lines.append(line)
                        svg = grid.svgs[0]
                        alllines.append(lines)
                        allconstructs_left.append(construct_left)
                        allconstructs_right.append(construct_right)
                        allsvgs.append(svg)
                        allkeys.append(grid.key())
                    except Exception, e:
                        pass
                self.set_message(type=u'error', content=u'Length: ' + str(len(alllines)))
                self.render(u'allgrids',
                    count = len(alllines),
                    count2=count2,
                    allkeys=allkeys,
                    allsvgString = allsvgs,
                    allconstructs_left = allconstructs_left, 
                    allconstructs_right = allconstructs_right, 
                    alllines=alllines, 
                    elements=elements)
            except Exception, e:
                self.set_message(type=u'error',
                    content=u'Unknown error occured. (' + unicode(e) + u') ' + unicode(traceback.print_exc()))






class SaveReflectionHandler(BaseHandler):
    csrf_protect = False
    @user_required
    def post(self):
        try:
            grids=Grid.find_by_user_ids([self.user.user_id], limit=1)
            grid = grids[0]
            reflection = self.request.POST[u'reflection']
            grid.reflections.append(db.Text(reflection))
            grid.reflectionwhens.append(datetime.datetime.now())
            grid.put()


            self.render(u'alldone',gridid=grid.key(),gridoffset=len(grid.svgs)-1)
        except Exception, e:
            self.set_message(type=u'error',
                content=u'Unknown error occured. (' + unicode(e) + u') ' + unicode(traceback.print_exc()))

class AdminHandler(BaseHandler):
    @user_required
    def get(self):
        self.render(u'admin')
        
class SaveAdminHandler(BaseHandler):
    @user_required
    def post(self):
        triads = self.request.POST[u'triads']
        elements = self.request.POST[u'elements']
        
        elementSingular = self.request.POST[u'elementSingular']
        elementPlural = self.request.POST[u'elementPlural']
        introblurb = self.request.POST[u'introblurb']
        thinkof = self.request.POST[u'thinkof']
        similarphrase = self.request.POST[u'similarphrase']
        differentphrase = self.request.POST[u'differentphrase']
        nowrate = self.request.POST[u'nowrate']
        wallid = self.request.POST[u'wallid']
        
        elementlist = elements.split('\r\n')
        TriadList.setTriadList(triads)
        Elements.setElements(elementlist)
        Configs.setConfigs(elementSingular,elementPlural,introblurb,thinkof,similarphrase,differentphrase,nowrate,wallid) 
        self.render(u'adminsaved')


def main():
    routes = [
        (r'/', RecentGridsHandler),
        (r'/user/(.*)', UserGridsHandler),
        (r'/run', GridHandler),
        (r'/grid', GridHandler),
        (r'/triad', TriadHandler),
        (r'/ratetriad', RatingsHandler),
        (r'/saveratings', SaveRatingsHandler),
        (r'/results', ResultsHandler),
        (r'/network', NetworkHandler),
        (r'/realtime', RealtimeHandler),
        (r'/showimage', ShowImageHandler),
        (r'/saveimage', SaveImageHandler),
        (r'/savereflection', SaveReflectionHandler),
        (r'/miserables', MiserablesHandler),
        (r'/allgrids', ShowAllHandler),
        (r'/admin', AdminHandler),
        (r'/saveadmin', SaveAdminHandler),
        
        (r'/task/refresh-user/(.*)', RefreshUserHandler)
    ]
    application = webapp.WSGIApplication(routes,
        debug=os.environ.get('SERVER_SOFTWARE', '').startswith('Dev'))
    util.run_wsgi_app(application)


if __name__ == u'__main__':
    main()
